'''
Created on Dec 29, 2020

@author: Alden
'''


# BED3, BED, BEDX, BEDGraph, BEDPE and all bed-derived file formats

from .baseio import BaseReader, BaseWriter
from .baseio import BaseIReader
from genomictools import GenomicAnnotation, StrandedGenomicAnnotation, GenomicPos, StrandedGenomicPos


class BED3(GenomicAnnotation):
	'''
	The basic BED3 format, with only chrom, chromStart and chromEnd 
	'''
	__slots__ = 'chrom', 'chromStart', 'chromEnd'
	def __init__(self, chrom, chromStart, chromEnd):
		self.chrom = chrom
		self.chromStart = chromStart
		self.chromEnd = chromEnd
# 		
	@property
	def genomic_pos(self):
		return GenomicPos(self.chrom, self.chromStart + 1, self.chromEnd)
	
	@staticmethod
	def from_genomic_pos(r):
		return BED3(r.name, r.zstart, r.ostop)

class BED3Reader(BaseReader):
	'''
	'''
	def _is_empty_field(self, s):
		return s == "" or s == "."
	def _proceed_next_line(self):
		while True:
			line = self.f.readline()
			if line == '':
				self._line = None
				break
			line = line.rstrip("\r\n") # Auto-stripping
			if line != '' and not line.startswith("#"): # Auto comment removal. Blank lines are skipped by default
				self._line = line
				break
	def _metainfo_reader(self):
		self._proceed_next_line()
		while self._line is not None and (self._line.startswith("track") or self._line.startswith("browser")):
			self._proceed_next_line()
	
	def _read(self):
		line = self._line
		if line is None:
			return None
		self._proceed_next_line() 
		words_array = line.split('\t')
		chrom = words_array[0]
		chromStart = int(words_array[1]) 
		chromEnd = int(words_array[2])
		return BED3(chrom, chromStart, chromEnd)
	
class BED3Writer(BaseWriter):
	def write(self, bed3):
		'''
		Output the BED3
		'''
		if isinstance(bed3, BED3):
			super(BED3Writer, self).write("{}\t{}\t{}\n".format(bed3.chrom, bed3.chromStart, bed3.chromEnd))
		else: # assume GenomicPos
			r = bed3.genomic_pos
			super(BED3Writer, self).write("{}\t{}\t{}\n".format(r.name, r.zstart, r.ostop))


		
	
class BED(BED3, StrandedGenomicAnnotation):
	'''
	The standard 12-field BED format 
	'''
	__slots__ = "name", "score", "strand", "thickStart", "thickEnd", "itemRgb", "blockCount", "blockSizes", "blockStarts"
	def __init__(self, chrom, chromStart, chromEnd, name=None, score=None, strand=None, thickStart=None, thickEnd=None, itemRgb=None, blockCount=None, blockSizes=None, blockStarts=None):
		super(BED, self).__init__(chrom, chromStart, chromEnd)
		self.name = name
		self.score = score
		self.strand = strand
		self.thickStart = thickStart
		self.thickEnd = thickEnd
		self.itemRgb = itemRgb
		self.blockCount = blockCount
		self.blockSizes = blockSizes
		self.blockStarts = blockStarts
	
	@property
	def stranded_genomic_pos(self):
		return StrandedGenomicPos(self.chrom, self.chromStart + 1, self.chromEnd, "." if self.strand is None else self.strand)
class BEDReader(BED3Reader):
	
	def _is_empty_field(self, s):
		return s is None or s == "" or s == "."
	
	def _read(self):
		line = self._line
		if line is None:
			return None
		self._proceed_next_line() 
		words_array = line.split('\t')
		chrom = words_array[0]
		chromStart = int(words_array[1]) 
		chromEnd = int(words_array[2])
		words_array.extend([None] * (12 - len(words_array)))
		
		name = words_array[3]
		score = None
		strand = None
		thickStart = None
		thickEnd = None
		itemRgb = None
		blockCount = None
		blockSizes = None
		blockStarts = None
		
		if not self._is_empty_field(words_array[4]):
			score = float(words_array[4])
		if words_array[5] != "":
			strand = words_array[5]
		if not self._is_empty_field(words_array[6]):
			thickStart = words_array[6]
		if not self._is_empty_field(words_array[7]):
			thickEnd = words_array[7]
		if not self._is_empty_field(words_array[8]):
			itemRgb = words_array[8]
		if not self._is_empty_field(words_array[9]):
			blockCount = int(words_array[9])
		if not self._is_empty_field(words_array[10]):
			blockSizes = list(map(int, filter(lambda a:a, words_array[10].split(","))))
		if not self._is_empty_field(words_array[11]):
			blockStarts = list(map(int, filter(lambda a:a, words_array[11].split(","))))
		
		if blockSizes is not None:
			if blockCount != len(blockSizes):
				raise Exception("Inconsistent blockCount and blockSizes")
			if blockCount != len(blockStarts):
				raise Exception("Inconsistent blockCount and blockStarts")
			if blockStarts[0] != 0:
				raise Exception("First block must start at 0")
			if chromStart + blockStarts[-1] + blockSizes[-1] != chromEnd:
				raise Exception("Last block must end at chromEnd")		 
		
		return BED(chrom, chromStart, chromEnd, name, score, strand, thickStart, thickEnd, itemRgb, blockCount, blockSizes, blockStarts)

	
	
class BEDWriter(BaseWriter):
	def write(self, bed):
		'''
		Output the BEDNode
		'''
		super(BEDWriter, self).write("{}\t{}\t{}\t{}\t{}\t{}\t{}\t{}\t{}\t{}\t{}\t{}\n".format(bed.chrom, bed.chromStart, bed.chromEnd, "" if bed.name is None else bed.name, "" if bed.score is None else bed.score, "" if bed.strand is None else bed.strand, "" if bed.thickStart is None else bed.thickStart, "" if bed.thickEnd is None else bed.thickEnd, "" if bed.itemRgb is None else bed.itemRgb, "" if bed.blockCount is None else bed.blockCount, "" if bed.blockSizes is None else ",".join(map(str, bed.blockSizes)), "" if bed.blockStarts is None else ",".join(map(str, bed.blockStarts))))


_bed_additional_fields = ["name", "score", "strand", "thickStart", "thickEnd", "itemRgb", "blockCount", "blockSizes", "blockStarts"]
def _block_func(s):
	return list(map(int, filter(lambda a:a, s.split(","))))
_bed_additional_field_funcs = [str, float, str, int, int, str, int, _block_func, _block_func] 		
class BEDXReader(BaseReader):
	'''
	A general BED-X class for customized BED file that extends the BEDx+ format.
	By default, input file is in bed3+ format. However, one could also apply this on bed6+, bed9+, etc.
	'''
	__slots__ = "fieldnames", "fieldfuncs", "BEDX"
	def __init__(self, arg, fieldnames, fieldfuncs=None, x=3, classname="BEDX"):
		'''
		fieldnames: List of names of the additional fields. 
		fieldfuncs: It can either be a list or a dict. 
		x: bedx+ 
		'''
		super(BEDXReader, self).__init__(arg)
		BEDXReaderSelf = self
		def _init(self, chrom, chromStart, chromEnd, *args):
			super(BEDXReaderSelf.BEDX, self).__init__(chrom, chromStart, chromEnd)
			if len(args) != len(self.__slots__):
				raise Exception("Inconsistent field names and entries")
			for slot, arg in zip(self.__slots__, args):
				setattr(self, slot, arg)
		if x < 3 or x > 12:
			raise Exception("Incorrect X")
		self.fieldnames = _bed_additional_fields[:x - 3] + fieldnames
		addition_field_funcs = {_bed_additional_fields[i]:_bed_additional_field_funcs[i] for i in range(x - 3)}
		if fieldfuncs is None:
			self.fieldfuncs = addition_field_funcs
		elif type(fieldfuncs) is dict:
			self.fieldfuncs = fieldfuncs
		else:
			self.fieldfuncs = _bed_additional_field_funcs[:x - 3] + fieldfuncs
		self.BEDX = type(classname, (BED3,), {"__slots__":self.fieldnames, "__init__":_init})
		
	def _read(self):
		line = self._line
		if line is None:
			return None
		self._proceed_next_line()
		words_array = line.split('\t')
		chrom = words_array[0]
		chromStart = int(words_array[1]) 
		chromEnd = int(words_array[2])
		if len(self.fieldnames) != len(words_array)-3:
			raise Exception("Number of fields in BED file does not match the number of field names")
		if self.fieldfuncs is None:
			return self.BEDX(chrom, chromStart, chromEnd, *words_array[3:])
		else:
			if type(self.fieldfuncs) is dict:
				args = [self.fieldfuncs[fieldname](obj) if fieldname in self.fieldfuncs else obj for obj, fieldname in zip(words_array[3:], self.fieldnames)]
			else:
				args = [fieldfunc(obj) for obj, fieldfunc in zip(words_array[3:], self.fieldfuncs)] 
			return self.BEDX(chrom, chromStart, chromEnd, *args)


class BEDGraph(BED3):
	__slots__ = "dataValue",
	def __init__(self, chrom, chromStart, chromEnd, dataValue):
		super(BEDGraph, self).__init__(chrom, chromStart, chromEnd)
		self.dataValue = dataValue
		
class BEDGraphReader(BaseReader):
	
	def __init__(self, arg, dataValueType=float):
		super(BEDGraphReader, self).__init__(arg)
		self.dataValueType = dataValueType
	def _metainfo_reader(self):
		self._proceed_next_line()
		while self._line is not None and (self._line.startswith("track") or self._line.startswith("browser")):
			self._proceed_next_line()
	
	def _read(self):
		line = self._line
		if line is None:
			return None
		self._proceed_next_line() 
		words_array = line.split('\t')
		chrom = words_array[0]
		chromStart = int(words_array[1]) 
		chromEnd = int(words_array[2])
		dataValue = self.dataValueType(words_array[3])
		return BEDGraph(chrom, chromStart, chromEnd, dataValue)
	
class BEDGraphWriter(BaseWriter):
	def __init__(self, arg, dataValueFunc=None):
		super(BEDGraphWriter, self).__init__(arg)
		if dataValueFunc is None:
			dataValueFunc = str
		self.dataValueFunc = dataValueFunc
		
	def write(self, bedgraph):
		'''
		Output the BEDGraph
		'''
		super(BEDGraphWriter, self).write("{}\t{}\t{}\t{}\n".format(bedgraph.chrom, bedgraph.chromStart, bedgraph.chromEnd, self.dataValueFunc(bedgraph.dataValue)))
	


class BEDPE():
	'''
	'''
	def __init__(self, chrom1, start1, stop1, chrom2, start2, stop2, name=None, score=None, strand1=None, strand2=None):
		self.chrom1 = chrom1
		self.start1 = start1
		self.stop1 = stop1
		self.chrom2 = chrom2
		self.start2 = start2
		self.stop2 = stop2
		self.name = name
		self.score = score
		self.strand1 = strand1
		self.strand2 = strand2
	@property
	def genomic_pos1(self):
		return GenomicPos(self.chrom1, self.start1 + 1, self.stop1)
	@property
	def stranded_genomic_pos1(self):
		return StrandedGenomicPos(self.chrom1, self.start1 + 1, self.stop1, self.strand1)
	@property
	def genomic_pos2(self):
		return GenomicPos(self.chrom2, self.start2 + 1, self.stop2)
	@property
	def stranded_genomic_pos2(self):
		return StrandedGenomicPos(self.chrom2, self.start2 + 1, self.stop2, self.strand2)
	
class BEDPEReader(BaseReader):
	def __init__(self, arg):
		super(BEDPEReader, self).__init__(arg)
	def _proceed_next_line(self):
		while True:
			line = self.f.readline()
			if line == '':
				self._line = None
				break
			line = line.rstrip("\r\n") # Auto-stripping
			if line != '' and not line.startswith("#"): # Auto comment removal. Blank lines are skipped by default
				self._line = line
				break
			
	def _read(self):
		if self._line is None:
			return None
		words_array = self._line.split('\t')
		self._proceed_next_line()
		name = None
		score = None
		strand1 = None
		strand2 = None
		chrom1 = words_array[0]
		start1 = int(words_array[1])
		stop1 = int(words_array[2])
		chrom2 = words_array[3]
		start2 = int(words_array[4])
		stop2 = int(words_array[5])
		if len(words_array) >= 7:
			name = words_array[6]
		if len(words_array) >= 8:
			score = words_array[7]
		if len(words_array) >= 9:
			strand1 = words_array[8]
		if len(words_array) >= 10:
			strand2 = words_array[9]
		return BEDPE(chrom1, start1, stop1, chrom2, start2, stop2, name, score, strand1, strand2)
		
	
	
class BEDPEWriter(BaseWriter):
	def __init__(self, arg):
		super(BEDPEWriter, self).__init__(arg)
			
	def write(self, bedpe):
		'''
		Output the BEDPENode
		Note that the start position is stored in 1-based but output in 0-based
		'''
		super(BEDPEWriter, self).write(
			"\t".join(list(map(str, [
				bedpe.chrom1, 
				bedpe.start1, 
				bedpe.stop1,
				bedpe.chrom2, 
				bedpe.start2, 
				bedpe.stop2, 
				"" if bedpe.name is None else bedpe.name, 
				"" if bedpe.score is None else bedpe.score, 
				"" if bedpe.strand1 is None else bedpe.strand1,
				"" if bedpe.strand2 is None else bedpe.strand2
				]))) + "\n")





'''
Created on Mar 31, 2021

@author: Alden
'''
import re

from .baseio import BaseReader, BaseWriter, get_text_file_extension
from collections import OrderedDict

def _guess_separator(f):
	# Auto separator detection based on file extension
	# This may be expanded in the future 
	file_extension = get_text_file_extension(f)
	if (file_extension == "tsv" or file_extension == "txt"): # Assume tab-delimited
		return "\t"
	elif (file_extension == "csv"): # Assume comma-delimited
		return ","
	else: 
		return "\t" # Default as tab-delimited

class DelimitedReader(BaseReader):
	'''
	A simple class for delimited file processing. Each line should have the same number of columns.
	
	This class targets for reading entry by entry (i.e. lines) instead of an entire table. 
	
	If the table has both column names and row names, and the header line contains exactly n_field_from_other_line - 1,
	 
	 
	One can define header by (1) setting header=True to automatically use the first non-comment line as header, or (2) using custom_header to define the header names. 
	The result object will be wrapped into a namedtuple  
	
	One can convert the element of each cell to specific type by defining custom_funcs.
	
	Example usage:
	
	.. code-block:: python
		
		import io
		from biodata.delimited import DelimitedReader
		
		print("===== Remove comment line =====")
		data = '# i am comment\n1\t2\n3\t4'
		print(data)
		print("-----------------------")
		with DelimitedReader(io.StringIO(data)) as dr:
			print(list(dr))
		with DelimitedReader(io.StringIO(data), comment_symbol="#", funcs=[int, int]) as dr:
			print(list(dr))
		print()
		print("===== Header =====")
		data = 'h1\th2\n1\t2\n3\t4'
		print(data)
		print("-----------------------")
		with DelimitedReader(io.StringIO(data)) as dr:
			print(list(dr))
		with DelimitedReader(io.StringIO(data), header=True) as dr:
			print(list(dr))
		with DelimitedReader(io.StringIO(data), header=True, funcs={"h1":int, "h2":int}) as dr:
			print(list(dr))
		print()
		print("===== header with comment symbol =====")
		data = '#h1\th2\n1\t2\n3\t4'
		print(data)
		print("-----------------------")
		with DelimitedReader(io.StringIO(data), header=True) as dr:
			print(list(dr))
		with DelimitedReader(io.StringIO(data), header=True, skip_header_comment_symbol="#") as dr:
			print(list(dr))
		print()
		print("===== Row name without header =====")
		data = 'r1\t1\t2\nr2\t3\t4\n'
		print(data)
		print("-----------------------")
		with DelimitedReader(io.StringIO(data)) as dr:
			print(list(dr))
		with DelimitedReader(io.StringIO(data), funcs=[str, int, int]) as dr:
			print(list(dr))
		with DelimitedReader(io.StringIO(data), funcs={1:int, 2:float}) as dr:
			print(list(dr))
		print()
		print("===== Row name with header =====")
		data = 'h1\th2\nr1\t1\t2\nr2\t3\t4'
		print(data)
		print("-----------------------")
		with DelimitedReader(io.StringIO(data)) as dr:
			print(list(dr))
		with DelimitedReader(io.StringIO(data), header=True, funcs={"h1":int, "h2":int}, additional_row_header="row_header") as dr:
			print(list(dr))
		print()
		
	.. code-block:: 
	
		===== Remove comment line =====
		# i am comment
		1	2
		3	4
		-----------------------
		[['# i am comment'], ['1', '2'], ['3', '4']]
		[[1, 2], [3, 4]]
		
		===== Header =====
		h1	h2
		1	2
		3	4
		-----------------------
		[['h1', 'h2'], ['1', '2'], ['3', '4']]
		[OrderedDict([('h1', '1'), ('h2', '2')]), OrderedDict([('h1', '3'), ('h2', '4')])]
		[OrderedDict([('h1', 1), ('h2', 2)]), OrderedDict([('h1', 3), ('h2', 4)])]
		
		===== header with comment symbol =====
		#h1	h2
		1	2
		3	4
		-----------------------
		[OrderedDict([('#h1', '1'), ('h2', '2')]), OrderedDict([('#h1', '3'), ('h2', '4')])]
		[OrderedDict([('h1', '1'), ('h2', '2')]), OrderedDict([('h1', '3'), ('h2', '4')])]
		
		===== Row name without header =====
		r1	1	2
		r2	3	4
		
		-----------------------
		[['r1', '1', '2'], ['r2', '3', '4']]
		[['r1', 1, 2], ['r2', 3, 4]]
		[['r1', 1, 2.0], ['r2', 3, 4.0]]
		
		===== Row name with header =====
		h1	h2
		r1	1	2
		r2	3	4
		-----------------------
		[['h1', 'h2'], ['r1', '1', '2'], ['r2', '3', '4']]
		[OrderedDict([('row_header', 'r1'), ('h1', 1), ('h2', 2)]), OrderedDict([('row_header', 'r2'), ('h1', 3), ('h2', 4)])]

	'''
	#__slots__ = "separator", "AutoDelimitedClass", "header_keys", "custom_funcs"
	
	def __init__(self, arg, *, separator=None, header=None, custom_header=None, funcs=None, comment_symbol=None, quote=None, skip_lines=None, skip_header_comment_symbol=None, additional_row_header=None):
		self.initialized = False
		super(DelimitedReader, self).__init__(arg)
		if isinstance(arg, str): 
			if separator is None:
				# Auto separator detection based on file extension
				self.separator = _guess_separator(arg)
			else:
				self.separator = separator
		else:
			if separator is None:
				self.separator = "\t" # By default use \t
			else:
				self.separator = separator
		
		# Raise exception for incorrect parameter settings	
		if header is not None and custom_header is not None:
			raise Exception()
		if header is None and skip_header_comment_symbol is not None:
			raise Exception()
		if header is None and additional_row_header is not None:
			raise Exception()
		
		self.header = header
		self.custom_header = custom_header
		self.funcs = funcs
		self.comment_symbol = comment_symbol
		self.quote = quote
		if quote is None:
			self.split_func = lambda line: line.split(self.separator)
		else: 
			self.split_func = lambda line: list(map("".join, re.findall(f'(?:(?<={self.separator})|(?<=^))(?:([^{self.quote}{self.separator}]*)|{self.quote}([^{self.quote}]*){self.quote})(?:(?={self.separator})|(?=$))', line)))
		#self.split_func = lambda line: list(csv.reader([line], delimiter=self.separator))[0]
		self.skip_lines = skip_lines
		self.skip_header_comment_symbol = skip_header_comment_symbol
		self.additional_row_header = additional_row_header
		self.initialized = True
		self._metainfo_reader()
		
			
	def _metainfo_reader(self):
		if not self.initialized:
			return
		if self.skip_lines is not None:
			for _ in range(self.skip_lines):
				self.f.readline()
		
		self._proceed_next_line()
		# Try to parse header
		self.header_keys = None
		if self.header is not None:
			line = self._line
			if self.skip_header_comment_symbol is not None:
				if not line.startswith(self.skip_header_comment_symbol):
					raise Exception()
				line = line[len(self.skip_header_comment_symbol):]
			words_array = self.split_func(line)
			if len(set(words_array)) != len(words_array):
				raise Exception()
			if self.additional_row_header is not None:
				self.header_keys = [self.additional_row_header] + words_array
			else:
				self.header_keys = words_array
			self._proceed_next_line()
			
		if self.custom_header is not None:
			self.header_keys = self.custom_header
		
			
		
	def _proceed_next_line(self):		
		'''
		Attempts to assign the next line to self._line. 
		'''
		while True:
			line = self.f.readline()
			if line == '':
				self._line = None
				break
			line = line.rstrip("\r\n") # Auto-stripping
			if line != '' and (self.comment_symbol is None or not line.startswith(self.comment_symbol)): # Auto comment removal. Blank lines are skipped by default
				self._line = line
				break
		
	def _read(self):
		if self._line is None:
			return None
		
		words_array = self.split_func(self._line)
		self._proceed_next_line()
		
		if self.header_keys is None:
			if self.funcs is not None:
				if type(self.funcs) is dict:
					words_array = [self.funcs[key](value) if key in self.funcs else value for key, value in enumerate(words_array)]
				else:
					if len(self.funcs) != len(words_array):
						raise Exception()
					words_array = [f(v) for v, f in zip(words_array, self.funcs)]
			return words_array
		else:
			if len(self.header_keys) != len(words_array):
				print(self.header_keys)
				print(words_array)
				raise Exception()
			if self.funcs is not None:
				words_array = [self.funcs[key](value) if key in self.funcs else value for key, value in zip(self.header_keys, words_array)]
			return OrderedDict((key, value) for key, value in zip(self.header_keys, words_array))


class DelimitedWriter(BaseWriter):
	'''
	A simple class for delimited file output.
	
	Example usage:
	
	.. code-block:: python
	
		with DelimitedWriter("Output.txt") as dw:
			dw.write(["A1", "A2", "A3"])
			dw.write(["B1", "B2", "B3"])
	'''
	__slots__ = "separator"
	def __init__(self, arg, separator=None):
		super(DelimitedWriter, self).__init__(arg)
		if separator is None:
			if isinstance(arg, str):
				self.separator = _guess_separator(arg)
			else:
				self.separator="\t" # Use \t
		else:
			self.separator = separator
			
	def write(self, arr):
		'''
		Output any iterable
		'''
		super(DelimitedWriter, self).write(self.separator.join(str(d) for d in arr) + "\n")
	
	
	
	


import sys
import bz2
import gzip
from io import TextIOBase
try:
	import Bio.bgzf
	_SUPPORT_BGZ = True
except:
	_SUPPORT_BGZ = False
class BaseReader(object):
	'''
	The base class for all sequential Readers handling different text file types. By default it (1) reads line-by-line, (2) skips blank lines, (3) rstrips the line.
	
	Any subclass should override the _read function. Other functions such as _proceed_next_line and _metainfo_reader should be overrided if applicable. 
	
	The file name '-' is treated as stdin. If a file ends in gz / bz2, it will use gzip / bz2 to open the file with utf-8 encoding

	Example usage:

	.. code-block:: python
		
		# The following are equivalent in reading a files into lines.
		from biodata.baseio import BaseReader
		with BaseReader(filename) as br:
			lines = br.read_all()
		
		from biodata.baseio import BaseReader
		lines = []
		with BaseReader(filename) as br:
			for b in br:
				lines.append(b)
		
		from biodata.baseio import BaseReader
		lines = BaseReader.read_all(list, filename)

		# TextIOBase can be used as input
		import io
		from biodata.baseio import BaseReader
		with BaseReader(io.StringIO("Line1\nLine2\n")) as br:
			lines = br.read_all()
		
	.. code-block:: python
	
		class ExampleNode(object):
			def __init__(self, value1, value2):
				self.value1 = value1
				self.value2 = value2
		
		class ExampleNodeReader(BaseReader):
			def __init__(self, filename):
				super(ExampleNodeReader, self).__init__(filename)
			def _read(self):
				if self.line is None:
					return None
				words_array = self.line.split('\\t')
				value1 = words_array[0]
				value2 = words_array[1]
				self.proceed_next_line()
				return ExampleNode(value1, value2)
		
		filename = "SomeDocument.txt"
		with ExampleNodeReader(filename) as er:
			node = er.read()
			while node is not None:
				print(node.value1 + "\\t" + node.value2)
				node = er.read()
	
	
	'''
	
	__slots__ = "f", "_line", "_is_stored", "_stored_value"
	def __init__(self, arg):
		'''
		Either an arg or 
		'''
		if isinstance(arg, TextIOBase):
			self.f = arg 
		else:
			self.f = create_text_stream(arg, "r")
		self._is_stored = False
		self._metainfo_reader()
			
		
	def _proceed_next_line(self):
		'''
		Attempts to assign the next line to self._line. 
		It is recommended for any subclass to override and use this method on how to proceed to next line (e.g. Ignore comment line starting with #) 	
		'''
		while True:
			line = self.f.readline()
			if line == '':
				self._line = None
				break
			line = line.rstrip("\r\n") # Auto-stripping
			if line != '': # Auto comment removal. Blank lines are skipped by default
				self._line = line
				break
			
	
	def _metainfo_reader(self):
		'''
		Attempts to process meta info in the data file. 
		Subclasses can override this method to skip headers with other formats or process the comment / header lines (e.g. file format version)
		'''
		self._proceed_next_line()
	
	
	def _read(self):
		line = self._line
		self._proceed_next_line()
		return line
	
	def read(self):
		'''
		Returns the next object
		'''
		if self._is_stored:
			self._is_stored = False
			return self._stored_value
		else:
			return self._read()
	
	def peek(self):
		if not self._is_stored:
			self._is_stored = True 
			self._stored_value = self._read()
		return self._stored_value 
	
	def __iter__(self):
		'''
		Create a generator for reading data
		'''
		obj = self.read()
		while obj is not None:
			yield obj
			obj = self.read()
			
	def close(self):
		'''
		Close the file
		'''
		if self.f is not sys.stdin:
			self.f.close()
	
	def __enter__(self): 
		return self
	
	def __exit__(self, type, value, traceback): 
		self.close()
	
	@classmethod
	def read_all(cls, read_all_func, *args, **kwargs):
		with cls(*args, **kwargs) as br:
			return read_all_func(br)

class BaseWriter(object):
	'''
	The base text class for all Writers handling different text file types. By default it overwrites the target file.
	
	The file name '-' is treated as stdout. If a file ends in gz / bz2, it will use gzip / bz2 to open the file with utf-8 encoding
	
	Example usage:

	.. code-block:: python
	
		with BaseWriter(filename) as bw:
			bw.write("This is a sentence written to the specified file.")
		
	.. code-block:: python
		
		with BaseWriter(create_file_stream(filename, mode="a")) as bw:
			bw.write("This is a sentence appended to the specified file")
		
	.. code-block:: python
	
		class ExampleNode(object):
			def __init__(self, value1, value2):
				self.value1 = value1
				self.value2 = value2
		
		class ExampleNodeWriter(BaseWriter):
			def __init__(self, filename):
				super(ExampleNodeWriter, self).__init__(filename)
			def _initialize_header(self):
				super(ExampleNodeWriter, self).write("#Value1\\tValue2\\n")
			def write(self, node):
				super(ExampleNodeWriter, self).write(node.value1 + "\\t" + node.value2 + "\\n")
		
		node = ExampleNode("1", "This is the sentence to be written in the specified file.")
		filename = "SomeDocument.txt"
		with ExampleNodeWriter(filename) as ew:
			ew.write(node)
		
	'''
	
	__slots__ = "f"
	def __init__(self, arg, mode="w"):
		if isinstance(arg, TextIOBase):
			self.f = arg
		else:
			self.f = create_text_stream(arg, mode)
		self._initialize_header()
	
	def _initialize_header(self):
		'''
		This method is automatically called during initialization. Override this method to initialize and output the headers before writing the content.
		'''
		pass
	
	def write(self, s):
		'''
		Writes the string to the file
		'''
		self.f.write(s)
	
	def close(self):
		'''
		Closes the file
		'''
		if self.f is not sys.stdout:
			self.f.close()
		
	def __enter__(self): 
		return self
	
	def __exit__(self, type, value, traceback): 
		self.close()

	@classmethod
	def write_all(cls, iterable, arg, *args, **kwargs):
		with cls(arg, *args, **kwargs) as bw:
			for i in iterable:
				bw.write(i)

class NullWriter(object):
	'''
	Creates a NULL writer, that does nothing when the user called the method write. This writer class is useful for optional output file. 
	

	Example usage:

	.. code-block:: python
	
		filename = None
		if filename is not None:
			bw = BaseWriter(filename)
		else:
			bw = NullWriter()
		bw.write("If filename is provided, this sentence will be written to the file. Otherwise, it will be written to nowhere")
		bw.close()
		
	'''
	def __init__(self):
		pass

	def write(self, s):
		pass
	
	def write_all(self, elements):
		pass
			
	def close(self):
		pass
		
	def __enter__(self): 
		return self
	
	def __exit__(self, type, value, traceback): 
		self.close()


def get_text_file_extension(filename, neglect_zip=True):
	'''
	Neglect the zip extension and return file extension 'txt', 'csv', etc.
	
	Only neglect one level of zip extension. "file.gz.bz2" will return 'gz'
	
	:param filename: The file name
	
	:returns: The file extension after decompression
	
	Example usage:
	
	.. code-block:: python
	
		get_text_file_extension("a.txt") # 'txt'
		
		get_text_file_extension("a.txt.gz") # 'txt'
		
		get_text_file_extension("a.csv.bz2") # 'csv'
		
		get_text_file_extension("a.txt.gz", neglect_zip = False) # 'gz'
		
		get_text_file_extension("example") # ''
	''' 
	supported_zip_extensions = [".gz", ".bz2"]
	if (neglect_zip):
		for supported_zip_extension in supported_zip_extensions:
			if filename.endswith(supported_zip_extension):
				filename = filename[0:filename.rfind(supported_zip_extension)]
				break
	
	index = filename.rfind(".")
	if index == -1: # No file extension found
		return ""
	else:
		return filename[index+1:] 
	

	
def create_text_stream(filename, mode):
	'''
	Create a text file stream of the file.
	
	Other than a file name, filename can also be:
	 
	(1) The filename "-" is interpreted as stdin or stdout.
	(2) The filename can be urllib.request.Request. 
	
	A decompressed stream is returned for compressed file (.gz or .bz2).
	
	All encodings are assumed utf-8    
	
	:param filename: The file name
	
	:returns: a TextIOBase representing the file stream
	
	Example usage:
	
	.. code-block:: python
		
		create_file_stream("file.txt", mode="r")
		create_file_stream(urllib.request.Request(url), mode="r")
	
	'''
	import urllib.request
	if isinstance(filename, urllib.request.Request):
		import io
		request = filename
		result = urllib.request.urlopen(request)
		remote_file_name = result.info().get_filename()
		stream = io.BytesIO(result.read())
		if remote_file_name.endswith(".bgz"):
			f = gzip.open(stream, mode = mode + 't', encoding='utf-8')
		elif remote_file_name.endswith(".gz"):
			f = gzip.open(stream, mode = mode + 't', encoding='utf-8')		
		elif remote_file_name.endswith(".bz2"):
			f = bz2.open(stream, mode = mode + 't', encoding='utf-8')
		else:
			f = io.TextIOWrapper(stream, encoding='utf-8')
	else:
		if filename == "-":
			if mode == 'r':
				f = sys.stdin
			else:
				f = sys.stdout
		else:
			if filename.endswith(".bgz"):
				if mode == "r":
					f = gzip.open(filename, mode = mode + 't', encoding='utf-8')
				else:
					if not _SUPPORT_BGZ:
						raise Exception("bgzip is not supported without Bio.bgzf")
					f = Bio.bgzf.open(filename, mode=mode+'t')
						
			elif filename.endswith(".gz"):
				f = gzip.open(filename, mode = mode + 't', encoding='utf-8')			
			elif filename.endswith(".bz2"):
				f = bz2.open(filename, mode = mode + 't', encoding='utf-8')
			else:
				f = open(filename, mode)
	return f


class BaseIReader(object):
	'''
	The base class for all readers of indexed-files. 
	For a typical file types that allows quick query onto certain entries.
	There is no guaranteed on what are indexed. 
	Also, there may not be any unique key per entry.
	Some files may have an additional associated index file.
	
	fasta, region-related, etc.  
	'''
	def __init__(self, arg):
		'''
		'''
		pass

	def __getitem__(self, key):
		# Exact or inexact
		pass
	
	def __enter__(self): 
		return self
	
	def __exit__(self, type, value, traceback): 
		self.close()



import csv
from collections import OrderedDict
import logging


from genomictools import StrandedGenomicAnnotation, StrandedGenomicPos
from .baseio import BaseReader, BaseWriter

def _quote_split(s, delimiter):
	return next(csv.reader([s],delimiter=delimiter))
		
	

__all__ = ["GFF3", "GFF3Reader", "GFF3Writer", "GTF", "GTFReader", "GTFWriter"]
class GFF3(StrandedGenomicAnnotation):
	def __init__(self, seqid, source, feature, start, end, score, strand, phase, attributes):
		super(GFF3, self).__init__()
		self.seqid = seqid
		self.source = source
		self.feature = feature
		self.start = start
		self.end = end
		self.score = score
		self.strand = strand
		self.phase = phase		
		self.attributes = attributes
	
	# For support to old GTF field name
	@property
	def seqname(self):
		return self.seqid	
	@property
	def attribute(self):
		return self.attributes
	@property
	def frame(self):
		return self.phase
	
	@property
	def stranded_genomic_pos(self):
		return StrandedGenomicPos(self.seqname, self.start, self.end, self.strand)
	
	
class GFF3Reader(BaseReader):
	def __init__(self, arg):
		if isinstance(arg, str):
			super(GFF3Reader, self).__init__(arg)
	def _metainfo_reader(self):
		line = self.f.readline()
		line = line.rstrip("\r\n") # Auto-stripping
		if line is not None:
			if line != "##gff-version 3":
				logging.warning("Missing GFF3-version header. This file may not be a proper GFF3 file.")
				# This warning may be modified in the future
			self._proceed_next_line()
		else:
			raise Exception("The first line must be ##gff-version 3.")
		
	def _proceed_next_line(self):
		'''
		Attempts to assign the next line to self._line. 
		'''
		while True:
			line = self.f.readline()
			if line == '':
				self._line = None
				break
			line = line.rstrip("\r\n") # Auto-stripping
			if line != '' and not line.startswith("#"): # Auto comment removal. Blank lines are skipped by default
				self._line = line
				break
		
	def _read(self):
		line = self._line
		if line is None:
			return None
		self._proceed_next_line()
		
		words_array = line.split('\t')
		seqid = words_array[0]
		source = words_array[1]
		feature = words_array[2]
		start = int(words_array[3])
		end = int(words_array[4])
		score = float(words_array[5]) if words_array[5] != "." else None
		strand = words_array[6]
		phase = words_array[7]
		attributes = OrderedDict([_quote_split(item.strip(), "=") for item in _quote_split(words_array[8], ";")])
		return GFF3(seqid, source, feature, start, end, score, strand, phase, attributes)
	@staticmethod
	def read_all_by_features(features, read_all_func, *args, **kwargs):
		'''
		A short cut method to read GFFs with selected features
		'''
		return GFF3Reader.read_all(lambda gff3_generator: read_all_func(filter(lambda gff3: gff3.feature in features, gff3_generator)), *args, **kwargs)

class GFF3Writer(BaseWriter):
	def __init__(self, arg):
		if isinstance(arg, str):
			super(GFF3Writer, self).__init__(arg)
	def _initialize_header(self):
		super(GFF3Writer, self).write("##gff-version 3\n")
	def write(self, gff3):
		'''
		Output the GFF
		Note that the start position is stored in 1-based
		'''
		super(GFF3Writer, self).write(("{}\t{}\t{}\t{}\t{}\t{}\t{}\t{}\t{}\n").format(
			gff3.seqid,
			gff3.source,
			gff3.feature,
			gff3.start,
			gff3.end,
			gff3.score if gff3.score is not None else ".",
			gff3.strand,
			gff3.phase,	
			";".join([key + "=" + value for key, value in gff3.attributes.items()])))
			


class GTF(StrandedGenomicAnnotation):
	# Equivalent to GFF2
	def __init__(self, seqname, source, feature, start, end, score, strand, frame, attribute):
		self.seqname = seqname
		self.source = source
		self.feature = feature
		self.start = start
		self.end = end
		self.score = score
		self.strand = strand
		self.frame = frame		
		self.attribute = attribute

	# For support to GFF3 field name
	@property
	def seqid(self):
		return self.seqname	
	@property
	def attributes(self):
		return self.attribute
	@property
	def phase(self):
		return self.frame
	
	@property
	def stranded_genomic_pos(self):
		return StrandedGenomicPos(self.seqname, self.start, self.end, self.strand)
class GTFReader(BaseReader):
	def _proceed_next_line(self):
		'''
		Attempts to assign the next line to self._line. 
		'''
		while True:
			line = self.f.readline()
			if line == '':
				self._line = None
				break
			line = line.rstrip("\r\n") # Auto-stripping
			if line != '' and not line.startswith("#"): # Auto comment removal. Blank lines are skipped by default
				self._line = line
				break
	
	def _read(self):
		line = self._line
		if line is None:
			return None
		self._proceed_next_line()
		words_array = line.split('\t')
		seqname = words_array[0]
		source = words_array[1]
		feature = words_array[2]
		start = int(words_array[3])
		end = int(words_array[4])
		score = float(words_array[5]) if words_array[5] != "." and words_array[5] != "None" else None
		strand = words_array[6]
		frame = words_array[7]
		attribute = OrderedDict([_quote_split(item.strip(), " ") for item in _quote_split(words_array[8], ";") if len(item.strip()) != 0]) # Defective
		return GTF(seqname, source, feature, start, end, score, strand, frame, attribute)
	

class GTFWriter(BaseWriter):
	def __init__(self, arg):
		super(GTFWriter, self).__init__(arg)
			
	def write(self, gtf):
		'''
		Output the GTFNode
		'''
		super(GTFWriter, self).write(("{}\t{}\t{}\t{}\t{}\t{}\t{}\t{}\t{}\n").format(
			gtf.seqname,
			gtf.source,
			gtf.feature,
			gtf.start,
			gtf.end,
			gtf.score if gtf.score is not None else ".",
			gtf.strand,
			gtf.frame,	
			"; ".join([key + " " + "\"" + value + "\"" for key, value in gtf.attribute.items()])))



from collections import OrderedDict
import re
import shlex
from .baseio import BaseReader

class MEMEAlphabetSymbol:
	def __init__(self, character, name, color):
		self.character = character
		self.name = name
		self.color = color
	
class MEMEAlphabet():
	def __init__(self, core_symbols, name=None, standard_like=None,
				ambig_symbols_dict={}, complements={}):
		self.core_symbols = core_symbols
		self.name = name
		self.standard_like = standard_like
		self.ambig_symbols_dict = ambig_symbols_dict
		self.complements = complements
	
	def __contains__(self, s):
		return s in self.core_symbols or s in self.ambig_symbols_dict
	
	def __len__(self):
		return len(self.core_symbols)
	
def _check_valid_alphabet_symbol(c):
	return re.match(r"^[A-Za-z0-9.\-*?]$", c) is not None
		
	
def _parse_alphabet(lines):
	'''
	Parse alphabet lines and return MEMEAlphabet
	'''
	def subparse(s):
		items = shlex.split(s)
		if len(items) > 1:
			character, name, color = items
			name = name[1:-1]
		else:
			character, name, color = items[0], None, None
		return MEMEAlphabetSymbol(character, name, color)
	
	name = None
	standard_like = None
	core_symbols = []
	complements = {}
	ambig_symbols_dict = {}
	for line in lines:
		line = line.strip()
		if line.startswith("#"):
			continue
		line = line.split("#")[0].strip()
		if len(line) == 0:
			continue
		if line.startswith("ALPHABET"):
			if "=" in line:
				for c in line.split("=")[-1].strip():
					core_symbols.append(MEMEAlphabetSymbol(c, None, None))
			else:
				name, standard_like = shlex.split(line)[1:]
				name = name[1:-1]
		elif "~" in line: # Complements ymbols
			tmp_entries = []
			for entry in line.split("~"):
				tmp_entries.append(subparse(entry.strip()))
			e1, e2 = tmp_entries
			complements[e1.character] = e2.character
			complements[e2.character] = e1.character
			core_symbols.extend(tmp_entries)
		elif "=" in line: # Ambiguous symbols
			k, v = line.split("=")
			ambig_symbols_dict[k.strip()] = v.strip()
		else:
			# Core symbols
			core_symbols.append(subparse(line.strip()))
	for symbol in core_symbols:
		if not _check_valid_alphabet_symbol(symbol.character):
			raise Exception(f"Invalid core symbol character - {symbol.character}")		
	for c in ambig_symbols_dict:
		if not _check_valid_alphabet_symbol(c):
			raise Exception("Invalid ambiguous symbol character")
	for characters in ambig_symbols_dict.values():
		for c in characters:
			if not any(c == symbol.character for symbol in core_symbols):
				raise Exception(f"Ambiguous symbol character matched to unknown core symbols - {c}")
	return MEMEAlphabet(core_symbols, name, standard_like,
				ambig_symbols_dict, complements)

	
	
	
class MEMEMotif():
	def __init__(self, identifier, alternate_name, matrix_dict, pwm, url):
		self.identifier = identifier
		self.alternate_name = alternate_name
		self.matrix_dict = matrix_dict
		self.pwm = pwm
		self.url = url
		
	@property	
	def motif_length(self):
		return len(self.pwm)
	@property
	def alphabet_length(self):
		return len(self.pwm[0])
	
class MEMEMotifReader(BaseReader):
	'''
	'''
	def _metainfo_reader(self):
		self._proceed_next_line()
		line = self._line
		meme_version = None
		alphabet = None
		strands = None
		background_letter_frequencies = None
		warnings = []
		while not line.startswith("MOTIF"):
			if line.startswith("MEME version"):
				meme_version = line.split("MEME version")[-1].strip()
				self._proceed_next_line()
			elif line.startswith("ALPHABET="):
				alphabet = _parse_alphabet([line])
				self._proceed_next_line()
			elif line.startswith("ALPHABET"):
				alphabet_lines = []
				while line != "END ALPHABET":
					alphabet_lines.append(line)
					self._proceed_next_line()
					line = self._line
				alphabet = _parse_alphabet(alphabet_lines)
				self._proceed_next_line()
			elif line.startswith("strands:"):
				strands = line.split("strands:")[-1].strip().split()
				self._proceed_next_line()
			elif line.startswith("Background letter frequencies"):
				pattern = re.compile(r"([^\s])\s(0\.[0-9]{3})")
				background_letter_frequencies = {}
				self._proceed_next_line()
				line = self._line
				while not line.startswith("MOTIF"):
					background_letter_frequencies.update({k:float(v) for k, v in pattern.findall(line)})
					self._proceed_next_line()
					line = self._line
				if sum(background_letter_frequencies.values()) != 1:
					warnings.append("Background letter frequencies not sum to 1")
				if alphabet is not None and len(background_letter_frequencies) != len(alphabet):
					warnings.append("Unmatched length of alphabet and background letter frequencies")
			else:
				self._proceed_next_line()
			line = self._line
		self.meme_version = meme_version
		self.alphabet = alphabet
		self.strands = strands
		self.background_letter_frequencies = background_letter_frequencies
		self.strands = strands
	
	def _proceed_next_line(self):
		while True:
			line = self.f.readline()
			if line == '':
				self._line = None
				break
			line = line.rstrip("\r\n") # Auto-stripping
			line = line.strip()
			if line != '': # Auto comment removal. Blank lines are skipped by default
				self._line = line
				break
	
	def _read(self):
		line = self._line
		if line is None:
			return None
		while not line.startswith("MOTIF"):
			self._proceed_next_line() 
			line = self._line
			if line is None:
				return None
		s1 = line.split()
		if len(s1) == 2:
			identifier = s1[1]
			alternate_name = None
		elif len(s1) == 3:
			identifier, alternate_name = s1[1:3]
		else:
			raise Exception("Invalid motif line - " + line)
		
		
		while not line.startswith("letter-probability matrix:"):
			self._proceed_next_line()
			line = self._line
		pattern = re.compile("\\s(\\S+)=\\s(\\S+)")
		matrix_dict = OrderedDict([i.groups() for i in pattern.finditer(line[26:])])
		funcs = {"alength": int, "w": int, "nsites": int, "E": float}
		matrix_dict = {k:(funcs[k](v) if k in funcs else v) for k, v in matrix_dict.items()}
		self._proceed_next_line()
		line = self._line
		
		pwm = []
		while line is not None and not line.startswith("URL") and not line.startswith("MOTIF"):
			pwm.append(list(map(float, line.split())))
			self._proceed_next_line()
			line = self._line
		if line.startswith("URL"):
			url = line.split()[1]
			self._proceed_next_line()
		else:
			url = None 
		return MEMEMotif(identifier, alternate_name, matrix_dict, pwm, url)



from setuptools import setup, find_packages

with open("README.md", "r") as readme_file:
	readme = readme_file.read()

requirements = ["genomictools>=0.0.5"]

setup(
	name="biodata",
	version="0.0.8",
	author="Alden Leung",
	author_email="alden.leung@gmail.com",
	description="A python package for common biological data I/O",
	long_description=readme,
	long_description_content_type="text/markdown",
	url="https://github.com/aldenleung/biodata/",
	packages=find_packages(),
	install_requires=requirements,
	classifiers=[
		"Programming Language :: Python :: 3.7",
		"Programming Language :: Python :: 3.8",
		"Programming Language :: Python :: 3.9",
		"Programming Language :: Python :: 3.10",
		"Programming Language :: Python :: 3.11",
		"License :: OSI Approved :: GNU General Public License v3 (GPLv3)",
	],
)



'''
Created on Dec 29, 2020

@author: Alden
'''
import textwrap
from .baseio import BaseReader, BaseWriter, BaseIReader
from genomictools import GenomicAnnotation, StrandedGenomicAnnotation

_reverse_complment_map = {
"A":"T",
"G":"C",
"C":"G",
"T":"A",
"Y":"R",
"R":"Y",
"W":"W",
"S":"S",
"K":"M",
"M":"K",
"D":"H",
"V":"B",
"H":"D",
"B":"V",
"X":"X",
"N":"N",
"a":"t",
"g":"c",
"c":"g",
"t":"a",
"y":"r",
"r":"y",
"w":"w",
"s":"s",
"k":"m",
"m":"k",
"d":"h",
"v":"b",
"h":"d",
"b":"v",
"x":"x",
"n":"n",
"-":"-"
}

def _reverse_complement(s):
	return "".join([_reverse_complment_map[c] for c in s][::-1])

class FASTA(object):
	'''
	The basic FASTA format
	'''
	
	def __init__(self, name, seq):
		self.name = name
		self.seq = seq
	
	def __len__(self):
		return len(self.seq)
	
class FASTAReader(BaseReader):
	'''
	'''
	def _read(self):
		if self._line is None:
			return None
		words_array = self._line.split('>')
		name = words_array[1]
		self._proceed_next_line()
		
		seqarray = []
		while self._line is not None and not self._line.startswith(">"):
			seqarray.append(self._line)
			self._proceed_next_line()	
		seq = "".join(seqarray)
				
		return FASTA(name, seq)
	
	
class Faidx():
	'''
	File format specification from htslib
	
	NAME	Name of this reference sequence
	LENGTH	Total length of this reference sequence, in bases
	OFFSET	Offset in the FASTA/FASTQ file of this sequence's first base
	LINEBASES	The number of bases on each line
	LINEWIDTH	The number of bytes in each line, including the newline
	QUALOFFSET	Offset of sequence's first quality within the FASTQ file
	'''
	def __init__(self, name, length, offset, linebases, linewidth, qualoffset):
		self.name = name
		self.length = length
		self.offset = offset
		self.linebases = linebases
		self.linewidth = linewidth
		self.qualoffset = qualoffset

class FaidxReader(BaseReader):
	def _read(self):
		if self._line is None:
			return None
		words_array = self._line.split()
		self._proceed_next_line()
		
		name = words_array[0]
		length = int(words_array[1])
		offset = int(words_array[2])
		linebases = int(words_array[3])
		linewidth = int(words_array[4])
		if len(words_array) > 5:
			qualoffset = int(words_array[5])
		else:
			qualoffset = None
		
		return Faidx(name, length, offset, linebases, linewidth, qualoffset)
	
class FASTAIReader(BaseIReader):
	def __init__(self, f, fai=None):
		if isinstance(f, str):
			if fai is None:
				fai = f + ".fai"
		if fai is None:
			raise Exception("Cannot auto-determine fai file")
		with FaidxReader(fai) as fr:
			self.faidx_dict = {idx.name:idx for idx in fr}
		self.f = open(f, "rb")
	def __getitem__(self, key):
		if isinstance(key, StrandedGenomicAnnotation):
			r = key.stranded_genomic_pos
			faidx = self.faidx_dict[r.name]
			if r.name not in self.faidx_dict:
				raise Exception("Seq name not found")
			faidx = self.faidx_dict[r.name]
			if r.zstart < 0 or r.ostop > faidx.length:
				raise Exception("Invalid range")
			start_offset = faidx.offset + r.zstart // faidx.linebases * faidx.linewidth + r.zstart % faidx.linebases
			stop_offset = faidx.offset + r.ostop // faidx.linebases * faidx.linewidth + r.ostop % faidx.linebases
			
			self.f.seek(start_offset)
			s = self.f.read(stop_offset - start_offset).decode()
			seq = "".join([c for c in s if c not in ["\r", "\n"]])
			if r.strand == "-":
				seq = _reverse_complement(seq)
			return FASTA(str(r), seq)
		elif isinstance(key, GenomicAnnotation):
			r = key.genomic_pos
			faidx = self.faidx_dict[r.name]
			if r.name not in self.faidx_dict:
				raise Exception("Seq name not found")
			faidx = self.faidx_dict[r.name]
			if r.zstart < 0 or r.ostop > faidx.length:
				raise Exception("Invalid range")
			start_offset = faidx.offset + r.zstart // faidx.linebases * faidx.linewidth + r.zstart % faidx.linebases
			stop_offset = faidx.offset + r.ostop // faidx.linebases * faidx.linewidth + r.ostop % faidx.linebases
			
			self.f.seek(start_offset)
			s = self.f.read(stop_offset - start_offset).decode()
			return FASTA(str(r), "".join([c for c in s if c not in ["\r", "\n"]]))			
		else:
			raise Exception("Unknown key format")
		
	def close(self):
		self.f.close()
		

class FASTAWriter(BaseWriter):
	def __init__(self, arg, maxcharac=-1):
		super(FASTAWriter, self).__init__(arg)
		self.maxcharac = maxcharac
	
	def write(self, fa):	
		self.f.write(">" + fa.name + "\n");
		if self.maxcharac > 0:
			s = "\n".join(textwrap.wrap(fa.seq, self.maxcharac, drop_whitespace=False, break_on_hyphens=False))
		else:
			s = fa.seq
		self.f.write(s + "\n");
		
class FASTQ(FASTA):
	def __init__(self, name, seq, quality):
		super(FASTQ, self).__init__(name, seq)
		if len(seq) != len(quality):
			raise Exception("Mismatch in length of sequence and quality")
		self.quality = quality

class FASTQReader(BaseReader):
	def _proceed_next_line(self):
		line = self.f.readline()
		if line == '':
			self._line = None
		else:
			line = line.rstrip("\r\n") # Auto-stripping
			self._line = line
			
	def _read(self):
		if self._line is None:
			return None
		words_array = self._line.split('@')
		name = words_array[1]
		self._proceed_next_line()
		seq = self._line		
		self._proceed_next_line()
		# "+"
		self._proceed_next_line()
		quality = self._line
		self._proceed_next_line()
		return FASTQ(name, seq, quality)
	

class FASTQWriter(BaseWriter):
	
	def initialize_header(self):
		pass
	
	def write(self, fa):	
		self.f.write("@" + fa.name + "\n");
		self.f.write(fa.seq + "\n");
		self.f.write("+\n");
		self.f.write(fa.quality + "\n");
	